/*
Title: SIXTIS
Author: Igor Galochkin, iseeall@narod.ru
Version: 1.3
Last update: 4 April 2006
*/
#include "sixtis.h"

//-----------------------------------------------------------------------------------------------------------------

void render (int dx, int dy)	
{// note: this function uses 10 for tile_size. Need to change all numbers here if blocks get reshaped
	int st_x (dx), st_y (dy), x (0), y (0), i (0);
	while (1)
	{
		switch (tile [i])
		{
			case BORDER: drawImagePart (image_blocks, st_x, st_y, 0, 0, 10, 10); break;
			case EMPTY: drawImagePart (image_blocks, st_x, st_y, 10, 0, 10, 10); break;
			case BLOCK: drawImagePart (image_blocks, st_x, st_y, 20, 0, 10, 10); break;
			case SHAPE: drawImagePart (image_blocks, st_x, st_y, 30, 0, 10, 10); break;
		}
		++i;
		++x;
		st_x += tile_size;
		if (x == field_w)
		{
			x = 0;
			st_x = dx;
			++y;
			st_y += tile_size;
			if (y == field_h - 1)	// don't draw last line - it will be drawn with the 'score: 0' string
				break;
		}
	}
	// draw GAME PAUSED message
	if (game_state == GAME_PAUSED)	
		//TextOut (hdc, 0, 0, "GAME PAUSED", 11);
		drawImagePart (image_words, 0, 0, 0, 30, 50, 10);
	

	// draw score
	if (field_w >= 12)
	{
		drawImagePart (image_words, st_x, st_y, 0, 0, 50, 10);	// 'score'
		st_x += 50;		
	}
	drawImagePart (image_blocks, st_x, st_y, 0,0,10,10);
	
	char s[20]; sprintf (s, "%d", score); // the number itself
	int n (0); st_x += 10;
	while (s[n])
	{
		char cc = s[n]; int c = cc - '0';		
		drawImagePart (image_words, st_x, st_y, (c%5)*10, (c/5)*10+10, 10, 10);
		++n; st_x += 10;		
	}	
	for (int i5 = (st_x - dx) / 10; i5 < field_w; ++i5, st_x += 10)	{	
		drawImagePart (image_blocks, st_x, st_y, 0,0,10,10);}
		// this thing here is needed to avoid blinking: if we first draw blocks and then draw score: 100 
		// over it, it blinks! So, the background images must be drawn only where there are no letters.
}

//----------------------------------------------------------------------------------------------------------------------

void input()
{
	if (GetAsyncKeyState(VK_RIGHT) < 0 ||
			GetAsyncKeyState('F') < 0 || // f
			GetAsyncKeyState('L') < 0 ) // l
				shape_abs_x += shifting_speed;	// shift right

	if (GetAsyncKeyState(VK_LEFT) < 0 ||
			GetAsyncKeyState('S') < 0 || // s
			GetAsyncKeyState('J') < 0 ) // j
				shape_abs_x -= shifting_speed;	// shift left

	if (GetAsyncKeyState(VK_DOWN) < 0 ||
			GetAsyncKeyState('D') < 0 || // d
			GetAsyncKeyState('K') < 0 ) // k
				shape_abs_y += fast_fall_speed;	// fast fall
}

//-----------------------------------------------------------------------------------------------------------------------

void rotate()	//rotates the shape
{ 
	clear_test_tiles();
	bool can_rotate = true;
	int x, y;
	for (y = shape_y; y != shape_y + shape_size; ++y)
	for (x = shape_x; x != shape_x + shape_size; ++x)
		if (in_field(x, y) && tile[x + y * field_w] == SHAPE)
		{
			int dest =	
				(shape_x + (y - shape_y)) + field_w *
					(shape_y + shape_size - 1 - (x - shape_x));
			(tile[dest] == BLOCK || tile[dest] == BORDER) ?
				can_rotate = false :
				test_tile[dest] = SHAPE;
		}				
	if (can_rotate)
		fetch_from_test_tile();	
}

//-----------------------------------------------------------------------------------------------------------------------

void check_rows()
{
// Function checks how many rows are filled, clears them up, and adds score
	int rows_full=0;
	vector<int>::iterator	rit = tile.begin() + field_w,
							rend = tile.end() - field_w;
	for(; rit != rend; rit += field_w)
	{
		bool row_full = true;
		vector<int>::iterator	row_it = rit,
								row_end = rit + field_w;
		for(; row_it != row_end; ++row_it)
		{
			if(*row_it == SHAPE)
				*row_it = BLOCK;		//convert shape to common blocks
			if(*row_it == EMPTY)
				row_full = false;		//row considered full if no empty spaces (so, if borders are not straight, rows may be of different width)
		}
		if(row_full)
		{
			vector<int>::iterator upper_it = row_it - field_w;
			for(; upper_it != tile.begin() - 1; --row_it, --upper_it)
				if(*row_it != BORDER)
				{
					*upper_it != BORDER ?
						*row_it = *upper_it :
						*row_it = EMPTY;
				}
			++rows_full;
		}
	}
	score += score_for_rows[rows_full];
}

//---------------------------------------------------------------------------------------------------------------

bool new_shape()
{
	// Creates new shape of random size at the top of the field
	// Number of blocks is random and can't exceed max_blocks_in_shape
	// extends of the shape can't exceed max_shape_size
	// If shape tries to appear over an already placed block, the game is over

	blocks_in_shape = random_int(max_blocks_in_shape - min_blocks_in_shape + 1) + min_blocks_in_shape;	
	shape_size = min (blocks_in_shape, max_shape_size); // limit size of the shape to max_shape_size

	shape_x = field_w / 2 - shape_size / 2 - 1; 
	if (shape_x < 1) shape_x = 1; // check needed for very small field size
	shape_y = 1;
	tile[field_w + field_w / 2 - 1] = SHAPE;  // put first block in the center and top of the field
	
	//descend_lag = 300+blocksSum*100; //speed of falling automatically becomes slower if the shape is large and faster if small
	
	int blocks_ready = 1;
	while (blocks_ready != blocks_in_shape)
	{
		bool can_place = false;
		int x = shape_x + (random_int(shape_size));
		int y = shape_y + (random_int(shape_size));
		int i = x + y * field_w;

		// seek if the tile will have filled neighbours so as to get sextamino, not just scattered blocks in 6x6 area
		int x2, y2;
		for(x2 = -1; x2 <= 1; ++x2)
			for(y2 = -1; y2 <= 1; ++y2)
				if(	in_field(x2 + x, y2 + y) && 
					tile[(x2 + x) + field_w * (y2 + y)] == SHAPE 
					&& (x2 * x2 * y2 * y2) != 1)		// diagonal neighbours don't count as neighbours
						can_place = true;
		if(can_place)
		{
			if(tile[i] == BLOCK)
				return false;							// CAN't PLACE NEW SHAPE - defeat condition
			else if(tile[i] == EMPTY)
			{ 
				tile[i] = SHAPE;	// put one more block of a new shape
				++blocks_ready;
			}
		}			
	}
	// reduce the shape_size in case real shape extends are
	// smaller than the maximal allowed (e.g. if shape has
	// 4 blocks, allowed size is 4, but if a 2x2 square is generated
	// we must reduce shape_size to 2 for smoother rotation)
	if (blocks_in_shape > 2)
	{
		int	x3 = shape_x,
				y3 = shape_y,					// running indices
				end_x3 = shape_x + shape_size,
				end_y3 = shape_y + shape_size,	// borders of area to search in (initial shape extends)
				st_shape_x = end_x3,
				st_shape_y = end_y3,			// where shape area will begin after correction
				end_shape_x = shape_x,
				end_shape_y = shape_y;			// where shape area will end after correction
		
		for(y3 = shape_y; y3 != end_y3; ++y3)
		for(x3 = shape_x; x3 != end_x3; ++x3)
			if(tile[x3 + field_w * y3] == SHAPE)
			{
				if(st_shape_x > x3)		st_shape_x = x3;
				if(st_shape_y > y3)		st_shape_y = y3;
				if(end_shape_x <= x3)	end_shape_x = x3 + 1;
				if(end_shape_y <= y3)	end_shape_y = y3 + 1; 
			}
		shape_x = st_shape_x;
		shape_y = st_shape_y;

		
		// if shape doesn't fit in a square, increase the smaller side
		// to make it fit, needed for smooth rotation
		int shape_w = end_shape_x - st_shape_x,
			shape_h = end_shape_y - st_shape_y;
		if(shape_w > shape_h)
		{
			int d = shape_w - shape_h;
			st_shape_y -= d/2;
			end_shape_y += (d - d/2); // not d/2, because d can be odd
			shape_h = shape_w;
		}
		if(shape_h > shape_w)
		{
			int d = shape_h - shape_w;
			st_shape_x -= d/2;
			end_shape_x += (d - d/2);
			shape_w = shape_h;
		}
		shape_size = shape_w;	
	}
	shape_abs_x = shape_x * tile_size;
	shape_abs_y = shape_y * tile_size;	
	return true;
}

//---------------------------------------------------------------------------

void game_start () // clears the field and put the first shape on the screen
{
	// update info about the field size and resize the game window
	field_size = field_w * field_h;	// update info on field size
	
	screen_width = max (field_w, 12) * tile_size;
	screen_height = max (field_h, 12) * tile_size;
	
	// auto-center the field in the window, if the field is smaller than 12 tiles in width/height
	camera_x = (field_w >= 12) ? 0 : (12 - field_w) * tile_size / 2; 
	camera_y = (field_h >= 12) ? 0 : (12 - field_h) * tile_size / 2;

	// Resize the window without sending any messages
	SetWindowPos(hwnd, NULL, 0, 0, 
							screen_width + 2 * GetSystemMetrics (SM_CXFIXEDFRAME), 
							screen_height + GetSystemMetrics (SM_CYCAPTION) + GetSystemMetrics (SM_CYMENU) + 2 * GetSystemMetrics (SM_CYFIXEDFRAME), 
							SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);

	vector<int> n (field_size, EMPTY);
	test_tile = tile = n;
	
	// create field borders
	for (int x = 0; x < field_w; ++x)	tile[x] = tile[x + (field_h - 1) * field_w] = BORDER;	
	for (int y = 0; y < field_h; ++y)	tile[y * field_w] = tile[field_w - 1 + y * field_w] = BORDER;

	score = 0;
	game_state = GAME_RUNNING;	
	new_shape();
}

//-----------------------------------------------------------------------------------------------------------------------

void game_end()
{  
	// Cleanup the device context and bitmaps
	DeleteObject (hdc);
	for (vector<HBITMAP>::size_type i = 0; i != loaded_images.size(); ++i)
		DeleteObject (loaded_images[i]);
	DestroyMenu (hmenu);
}

//---------------------------------------------------------------------------------------------------------

bool checkCollisions ()	// returns true if game is ended
{	
	int new_shape_x = static_cast<int>(shape_abs_x) / tile_size;
	int new_shape_y = static_cast<int>(shape_abs_y) / tile_size;
	
	int x, new_x, y, new_y, i, new_i;
	bool can_move;
	
	// check horizontal collision
	clear_test_tiles ();
	can_move = true;		
	for (x = shape_x, new_x = new_shape_x; can_move && x != shape_x + shape_size; ++x, ++new_x)
	for (y = shape_y; can_move && y != shape_y + shape_size; ++y)
	{
		i = y * field_w + x;
		new_i = i - x + new_x; // <=> new_i = y * field_w + new_x
		if (tile [i] == SHAPE)
			(tile [new_i] == BLOCK || tile [new_i] == BORDER) ?
				can_move = false :				
				test_tile[new_i] = SHAPE; 	
	}
	if (can_move)
	{
		shape_x = new_shape_x;	// move the shape right/left
		fetch_from_test_tile ();
	}
	else 
		shape_abs_x = shape_x * tile_size;	// restore position of the shape so that is doesn't overlap with obstacle
	
	// check vertical collision
	clear_test_tiles ();
	can_move = true;
	for (x = shape_x; can_move && x != shape_x + shape_size; ++x)
	for (y = shape_y, new_y = new_shape_y; can_move && y != shape_y + shape_size; ++y, ++new_y)
	{	
		i = y * field_w + x;
		new_i = new_y * field_w + x;
		if (tile [i] == SHAPE)
			(tile [new_i] == BLOCK || tile [new_i] == BORDER) ?
				can_move = false :				
				test_tile[new_i] = SHAPE; 	
	}	
	if(can_move)
	{
		shape_y = new_shape_y;	// move the shape down
		fetch_from_test_tile();			
	}
	else
	{ 
		check_rows();								// check if any full rows are placed
		return (!new_shape());			// try to place new shape and end game if it fails
	}
	return false;	// defeat condition not occured, continue game
}

//----------------------------------------------------------------------------------------------------------------------

void physics ()
{	
	shape_abs_y += falling_speed;
	if (checkCollisions ())
	{
		message ("Your score: %d", score);
		game_start ();				// restart the game
	}
}

//------------------------------------------------------------------------------------

BOOL CALLBACK AboutDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
		case WM_INITDIALOG:
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
				case IDOK:
				case IDCANCEL:
					EndDialog (hDlg, FALSE);
					return TRUE;
			}
		break;
	}
	return FALSE;
}

//------------------------------------------------------------------------------------
#define TO_INT(x) ((int)(x * 10.0))

BOOL CALLBACK NewGameDlgProc (HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static const int		min_field_w (4),						max_field_w (50),
											min_field_h (6),						max_field_h (100),
											min_max_blocks_in_shape (4),max_max_blocks_in_shape (25),
											min_min_blocks_in_shape (1),max_min_blocks_in_shape (10),
											min_max_shape_size (2),			max_max_shape_size (20);
	static const double min_falling_speed (0.1),		max_falling_speed (1.9),
											min_shifting_speed (2.0),		max_shifting_speed (9.0),
											min_fast_fall_speed (4.0),	max_fast_fall_speed (8.0);	
	static double	new_falling_speed (falling_speed),
								new_shifting_speed (shifting_speed),
								new_fast_fall_speed (fast_fall_speed);
	static int		new_field_w (field_w),
								new_field_h (field_h),
								new_max_blocks_in_shape (max_blocks_in_shape),
								new_min_blocks_in_shape (min_blocks_in_shape),
								new_max_shape_size (max_shape_size);
// sum of fast_fall_speed and fall_speed must not be more than 10 (tile_size), cuz the shape may jump thru an obstacle
					
	switch (msg)
	{
		case WM_INITDIALOG:	
		{
			HWND hFallingSpeedScroll = GetDlgItem (hDlg, IDC_FALLINGSPEED_SCROLL);
			SetScrollRange (hFallingSpeedScroll, SB_CTL, TO_INT(min_falling_speed), TO_INT (max_falling_speed), FALSE);
			SetScrollPos (hFallingSpeedScroll, SB_CTL, TO_INT(falling_speed), FALSE);
			char c1[20]; sprintf (c1, "%4.1f", new_falling_speed); 
			SetDlgItemText (hDlg, IDC_FALLINGSPEED, c1);
				
			HWND hShiftingSpeedScroll = GetDlgItem (hDlg, IDC_SHIFTINGSPEED_SCROLL);
			SetScrollRange (hShiftingSpeedScroll, SB_CTL, TO_INT(min_shifting_speed), TO_INT(max_shifting_speed), FALSE);
			SetScrollPos (hShiftingSpeedScroll, SB_CTL, TO_INT(shifting_speed), FALSE);
			char c2[20]; sprintf (c2, "%4.1f", new_shifting_speed); 
			SetDlgItemText (hDlg, IDC_SHIFTINGSPEED, c2);
			
			HWND hFastFallSpeedScroll = GetDlgItem (hDlg, IDC_FASTFALLSPEED_SCROLL);
			SetScrollRange (hFastFallSpeedScroll, SB_CTL, TO_INT(min_fast_fall_speed), TO_INT(max_fast_fall_speed), FALSE);
			SetScrollPos (hFastFallSpeedScroll, SB_CTL, TO_INT(fast_fall_speed), FALSE);
			char c3[20]; sprintf (c3, "%4.1f", new_fast_fall_speed); 
			SetDlgItemText (hDlg, IDC_FASTFALLSPEED, c3);
	
			HWND hFieldWidthScroll = GetDlgItem (hDlg, IDC_FIELDWIDTH_SCROLL);
			SetScrollRange (hFieldWidthScroll, SB_CTL, min_field_w, max_field_w, FALSE);
			SetScrollPos (hFieldWidthScroll, SB_CTL, new_field_w, FALSE);
			SetDlgItemInt (hDlg, IDC_FIELDWIDTH, new_field_w, FALSE);
			
			HWND hFieldHeightScroll = GetDlgItem (hDlg, IDC_FIELDHEIGHT_SCROLL);
			SetScrollRange (hFieldHeightScroll, SB_CTL, min_field_h, max_field_h, FALSE);
			SetScrollPos (hFieldHeightScroll, SB_CTL, new_field_h, FALSE);
			SetDlgItemInt (hDlg, IDC_FIELDHEIGHT, new_field_h, FALSE);
			
			HWND hMaxBlocksInShapeScroll = GetDlgItem (hDlg, IDC_MAXBLOCKSINSHAPE_SCROLL);
			SetScrollRange (hMaxBlocksInShapeScroll, SB_CTL, min_max_blocks_in_shape, max_max_blocks_in_shape, FALSE);
			SetScrollPos (hMaxBlocksInShapeScroll, SB_CTL, new_max_blocks_in_shape, FALSE);
			SetDlgItemInt (hDlg, IDC_MAXBLOCKSINSHAPE, new_max_blocks_in_shape, FALSE);
			
			HWND hMinBlocksInShapeScroll = GetDlgItem (hDlg, IDC_MINBLOCKSINSHAPE_SCROLL);
			SetScrollRange (hMinBlocksInShapeScroll, SB_CTL, min_min_blocks_in_shape, max_min_blocks_in_shape, FALSE);
			SetScrollPos (hMinBlocksInShapeScroll, SB_CTL, new_min_blocks_in_shape, FALSE);
			SetDlgItemInt (hDlg, IDC_MINBLOCKSINSHAPE, new_min_blocks_in_shape, FALSE);
			
			HWND hMaxShapeSizeScroll = GetDlgItem (hDlg, IDC_MAXSHAPESIZE_SCROLL);
			SetScrollRange (hMaxShapeSizeScroll, SB_CTL, min_max_shape_size, max_max_shape_size, FALSE);
			SetScrollPos (hMaxShapeSizeScroll, SB_CTL, new_max_shape_size, FALSE);
			SetDlgItemInt (hDlg, IDC_MAXSHAPESIZE, new_max_shape_size, FALSE);
		}	
		return TRUE;
		
		case WM_HSCROLL:
		{
			int control_id = GetWindowLong ((HWND)lParam, GWL_ID);
			switch (control_id)
			{						
				case IDC_FALLINGSPEED_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_falling_speed = max (min_falling_speed, new_falling_speed - 0.1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_falling_speed = min (max_falling_speed, new_falling_speed + 0.1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_falling_speed = (HIWORD (wParam)) / 10.0;													break;
					}	
					SetScrollPos (GetDlgItem (hDlg, IDC_FALLINGSPEED_SCROLL), SB_CTL, TO_INT(new_falling_speed), TRUE);
					char c1[20]; sprintf (c1, "%4.1f", new_falling_speed); 
					SetDlgItemText (hDlg, IDC_FALLINGSPEED, c1);
				break;
				
				case IDC_SHIFTINGSPEED_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_shifting_speed = max (min_shifting_speed, new_shifting_speed - 0.1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_shifting_speed = min (max_shifting_speed, new_shifting_speed + 0.1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_shifting_speed = (HIWORD (wParam)) / 10.0;														break;
					}	
					SetScrollPos (GetDlgItem (hDlg, IDC_SHIFTINGSPEED_SCROLL), SB_CTL, TO_INT(new_shifting_speed), TRUE);
					char c2[20]; sprintf (c2, "%4.1f", new_shifting_speed); 
					SetDlgItemText (hDlg, IDC_SHIFTINGSPEED, c2);
				break;
				
				case IDC_FASTFALLSPEED_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_fast_fall_speed = max (min_fast_fall_speed, new_fast_fall_speed - 0.1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_fast_fall_speed = min (max_fast_fall_speed, new_fast_fall_speed + 0.1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_fast_fall_speed = (HIWORD (wParam)) / 10.0;															break;
					}	
					SetScrollPos (GetDlgItem (hDlg, IDC_FASTFALLSPEED_SCROLL), SB_CTL, TO_INT(new_fast_fall_speed), TRUE);
					char c3[20]; sprintf (c3, "%4.1f", new_fast_fall_speed); 
					SetDlgItemText (hDlg, IDC_FASTFALLSPEED, c3);
				break;
				
				case IDC_FIELDWIDTH_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_field_w = max (min_field_w, new_field_w - 1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_field_w = min (max_field_w, new_field_w + 1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_field_w = HIWORD (wParam);										break;
					}	
					SetScrollPos (GetDlgItem (hDlg, IDC_FIELDWIDTH_SCROLL), SB_CTL, (new_field_w), TRUE);					
					SetDlgItemInt (hDlg, IDC_FIELDWIDTH, new_field_w, FALSE);
				break;
				
				case IDC_FIELDHEIGHT_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_field_h = max (min_field_h, new_field_h - 1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_field_h = min (max_field_h, new_field_h + 1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_field_h = HIWORD (wParam);										break;
					}	
					SetScrollPos (GetDlgItem (hDlg, IDC_FIELDHEIGHT_SCROLL), SB_CTL, (new_field_h), TRUE);					
					SetDlgItemInt (hDlg, IDC_FIELDHEIGHT, new_field_h, FALSE);
				break;
				
				case IDC_MAXBLOCKSINSHAPE_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_max_blocks_in_shape = max (min_max_blocks_in_shape, new_max_blocks_in_shape - 1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_max_blocks_in_shape = min (max_max_blocks_in_shape, new_max_blocks_in_shape + 1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_max_blocks_in_shape = HIWORD (wParam);										break;
					}	
				break;
				
				case IDC_MINBLOCKSINSHAPE_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_min_blocks_in_shape = max (min_min_blocks_in_shape, new_min_blocks_in_shape - 1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_min_blocks_in_shape = min (max_min_blocks_in_shape, new_min_blocks_in_shape + 1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_min_blocks_in_shape = HIWORD (wParam);										break;
					}	
				break;				
				
				case IDC_MAXSHAPESIZE_SCROLL:
					switch (LOWORD (wParam))
					{
						case SB_LINELEFT: case SB_PAGELEFT:						new_max_shape_size = max (min_max_shape_size, new_max_shape_size - 1);	break;
						case SB_LINERIGHT: case SB_PAGERIGHT:					new_max_shape_size = min (max_max_shape_size, new_max_shape_size + 1);	break;
						case SB_THUMBPOSITION: case SB_THUMBTRACK:		new_max_shape_size = HIWORD (wParam);										break;
					}	
				break;
			}	
			// correct settings
			// shape must fit into the field
			new_max_shape_size = min (new_max_shape_size, new_field_w - 2);
			new_max_shape_size = min (new_max_shape_size, new_field_h - 2);
			// shape can't have more blocks than square of its side (e.g. 4x4 shape can't have >16 blocks)
			new_max_blocks_in_shape = min (new_max_blocks_in_shape, new_max_shape_size * new_max_shape_size);
			// min number of blocks can't be higher than max number ))
			new_min_blocks_in_shape = min (new_min_blocks_in_shape, new_max_blocks_in_shape);
			
			SetScrollPos (GetDlgItem (hDlg, IDC_MAXSHAPESIZE_SCROLL), SB_CTL, (new_max_shape_size), TRUE);					
			SetDlgItemInt (hDlg, IDC_MAXSHAPESIZE, new_max_shape_size, FALSE);
				
			SetScrollPos (GetDlgItem (hDlg, IDC_MAXBLOCKSINSHAPE_SCROLL), SB_CTL, (new_max_blocks_in_shape), TRUE);					
			SetDlgItemInt (hDlg, IDC_MAXBLOCKSINSHAPE, new_max_blocks_in_shape, FALSE);		
					
			SetScrollPos (GetDlgItem (hDlg, IDC_MINBLOCKSINSHAPE_SCROLL), SB_CTL, (new_min_blocks_in_shape), TRUE);					
			SetDlgItemInt (hDlg, IDC_MINBLOCKSINSHAPE, new_min_blocks_in_shape, FALSE);	
		}	
		return FALSE;
		
		case WM_COMMAND:
			switch (LOWORD (wParam))
			{
				case IDCANCEL:
				// restore old values for parameters
						new_falling_speed = falling_speed;
						new_shifting_speed = shifting_speed;
						new_fast_fall_speed = fast_fall_speed;
						new_field_w = field_w;
						new_field_h = field_h;
						new_max_blocks_in_shape = max_blocks_in_shape;
						new_min_blocks_in_shape = min_blocks_in_shape;
						new_max_shape_size = max_shape_size;
					EndDialog (hDlg, FALSE);
				return TRUE;
				
				case IDOK:
				// apply changes
					falling_speed = new_falling_speed;
					shifting_speed = new_shifting_speed;
					fast_fall_speed = new_fast_fall_speed;
					field_w = new_field_w;
					field_h = new_field_h;
					max_blocks_in_shape = new_max_blocks_in_shape;
					min_blocks_in_shape = new_min_blocks_in_shape;
					max_shape_size = new_max_shape_size;
					
					game_start ();
					EndDialog (hDlg, FALSE);
				return TRUE;
			}
		break;
	}
	return FALSE;
}

//------------------------------------------------------------------------------------
LRESULT CALLBACK WndProc (HWND hWindow, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch (msg)
	{
		case WM_CREATE:
			game_start (); 
			image_blocks = loadImage ("BLOCKS");
			image_words = loadImage ("WORDS");
		return 0;

		case WM_ACTIVATE:
			if (wParam == WA_INACTIVE)	game_state =  GAME_PAUSED;
			repaint (true);	
		return 0;
			
		case WM_SIZE:
			if (wParam == SIZE_MINIMIZED) 	game_state = GAME_PAUSED;
			//message ("window resized!");
		break;
		
		case WM_PAINT:
			PAINTSTRUCT ps;
			hdc = BeginPaint(hWindow, &ps);
			render (camera_x, camera_y);
			EndPaint(hWindow, &ps);
			return 0;
		
		case WM_KEYDOWN:
			if (game_state == GAME_PAUSED)
				game_state = GAME_RUNNING;
			else
				switch (wParam)
				{
					case VK_UP: 
					case VK_SPACE:
					case ('E'): // e
					case ('I'): // i
						if (game_state == GAME_RUNNING)
							rotate();									// rotate shape 
					break;
					case VK_RETURN:
					case VK_ESCAPE: 
					case VK_PAUSE:
					case ('P'): // p
					case ('Q'): // q
						game_state = GAME_PAUSED;		// pause
						repaint (true);
					break;
				}
			return 0;
		
		case WM_LBUTTONDOWN:	// click pauses / unpauses game
			if (game_state == GAME_RUNNING) game_state = GAME_PAUSED;
			else if (game_state == GAME_PAUSED) game_state = GAME_RUNNING;
			repaint (true);
		return 0;
		
		case WM_COMMAND:	
			switch (LOWORD (wParam))
			{
				case ID_GAME_RESTART:	game_start (); break;
				case ID_GAME_NEWGAME: 				
					DialogBox (hinstance, TEXT ("NEWGAME"), hWindow, NewGameDlgProc); break;
				case ID_GAME_EXIT: SendMessage (hWindow, WM_CLOSE, 0, 0); break;
				
				case ID_EXTRA_STANDARDTETRIS:	field_w = 12; field_h = 27; max_blocks_in_shape = 4; min_blocks_in_shape = 4; max_shape_size = 4; falling_speed = 1.5; game_start (); break;
				case ID_EASY_BEGINNERTRIS:		field_w = 10; field_h = 20; max_blocks_in_shape = 4;  min_blocks_in_shape = 1; max_shape_size = 4; falling_speed = 0.5; game_start (); break;
				case ID_EASY_MINITETRIS:			field_w = 6; field_h = 15; max_blocks_in_shape = 4;  min_blocks_in_shape = 3; max_shape_size = 4; falling_speed = 0.5; game_start (); break;
				case ID_EASY_MICROTETRIS:			field_w = 4; field_h = 6; max_blocks_in_shape = 4;  min_blocks_in_shape = 2; max_shape_size = 2; falling_speed = 0.3; game_start (); break;
				case ID_EASY_MEGATETRIS:			field_w = 30; field_h = 30; max_blocks_in_shape = 4;  min_blocks_in_shape = 4; max_shape_size = 4; falling_speed = 1.0; game_start (); break;
				case ID_EASY_PENTIS:					field_w = 16; field_h = 30; max_blocks_in_shape = 5;  min_blocks_in_shape = 4; max_shape_size = 5; falling_speed = 0.8; game_start (); break;
				case ID_EASY_SIXTIS:					field_w = 18; field_h = 30; max_blocks_in_shape = 6;  min_blocks_in_shape = 3; max_shape_size = 6; falling_speed = 0.6; game_start (); break;
				case ID_EASY_FASTFINGERS:			field_w = 10; field_h = 14; max_blocks_in_shape = 3;  min_blocks_in_shape = 1; max_shape_size = 3; falling_speed = 1.9; game_start (); break;
				case ID_HARD_OCTIS:						field_w = 20; field_h = 32; max_blocks_in_shape = 8;  min_blocks_in_shape = 3; max_shape_size = 8; falling_speed = 0.4; game_start (); break;
				case ID_HARD_12IS:						field_w = 24; field_h = 36; max_blocks_in_shape = 12; min_blocks_in_shape = 2; max_shape_size = 12; falling_speed = 0.3; game_start (); break;
				case ID_HARD_16IS:						field_w = 30; field_h = 40; max_blocks_in_shape = 16; min_blocks_in_shape = 1; max_shape_size = 16; falling_speed = 0.2; game_start (); break;
				case ID_HARD_WELL:						field_w = 8;	field_h = 40; max_blocks_in_shape = 8; min_blocks_in_shape = 1; max_shape_size = 6; falling_speed = 0.8; game_start (); break;
				case ID_HARD_SEABOTTOM:				field_w = 50;	field_h = 10; max_blocks_in_shape = 5; min_blocks_in_shape = 1; max_shape_size = 5; falling_speed = 0.4; game_start (); break;
				case ID_HARD_RUSH:						field_w = 16;	field_h = 20; max_blocks_in_shape = 6; min_blocks_in_shape = 1; max_shape_size = 8; falling_speed = 1.9; game_start (); break;
				case ID_INSANE_MONSTERS:			field_w = 30; field_h = 40; max_blocks_in_shape = 25; min_blocks_in_shape = 1; max_shape_size = 20; falling_speed = 0.1; game_start (); break;
				case ID_INSANE_MONSTERSLOBBY:	field_w = 18; field_h = 40; max_blocks_in_shape = 16; min_blocks_in_shape = 6; max_shape_size = 4; falling_speed = 0.2; game_start (); break;
				case ID_INSANE_INCOMING:			field_w = 20; field_h = 30; max_blocks_in_shape = 25; min_blocks_in_shape = 6; max_shape_size = 5; falling_speed = 0.1; game_start (); break;
				case ID_INSANE_TURBOFINGERS:	field_w = 12; field_h = 20; max_blocks_in_shape = 12; min_blocks_in_shape = 1; max_shape_size = 6; falling_speed = 1.9; game_start (); break;
				case ID_INSANE_SINKINGSHIPS:	field_w = 40; field_h = 20; max_blocks_in_shape = 12; min_blocks_in_shape = 4; max_shape_size = 6; falling_speed = 0.4; game_start (); break;
				case ID_INSANE_UNDERPRESSURE:	field_w = 40; field_h = 6; max_blocks_in_shape = 4; min_blocks_in_shape = 1; max_shape_size = 4; falling_speed = 0.3; game_start (); break;
				case ID_INSANE_RECYCLEBIN:		field_w = 8; field_h = 24; max_blocks_in_shape = 16; min_blocks_in_shape = 1; max_shape_size = 4; falling_speed = 0.2; game_start (); break;
				case ID__ABOUT:
					DialogBox (hinstance, TEXT ("ABOUT"), hWindow, AboutDlgProc); break;
			}			
		return 0;
		
		case WM_DESTROY:// End the game and exit the application		
			game_end();
			PostQuitMessage (0);
			return 0;
	}
	return DefWindowProc(hWindow, msg, wParam, lParam);
}

//------------------------------------------------------------------------------------

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow)
{
	int cur_tick (0), next_repaint_tick (0);
	
// window initialization routine
	MSG         msg;
	WNDCLASS    wndclass;
	wndclass.style          = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc    = WndProc;
	wndclass.cbClsExtra     = 0;
	wndclass.cbWndExtra     = 0;
	wndclass.hInstance      = hInstance;
	hinstance = hInstance;
	wndclass.hIcon          = LoadIcon(hinstance, game_title);
	wndclass.hCursor        = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground  = (HBRUSH)GetStockObject (GRAY_BRUSH);;
	wndclass.lpszMenuName   = game_title;
	wndclass.lpszClassName  = game_title;

	if (!RegisterClass(&wndclass))
		return 0;

	hwnd = CreateWindow(game_title, game_title, 							
							WS_POPUPWINDOW | WS_CAPTION | WS_MINIMIZEBOX, 
							CW_USEDEFAULT, CW_USEDEFAULT, 
							screen_width + 2 * GetSystemMetrics (SM_CXFIXEDFRAME), 
							screen_height + GetSystemMetrics (SM_CYCAPTION) + GetSystemMetrics (SM_CYMENU) + 2 * GetSystemMetrics (SM_CYFIXEDFRAME), 
							NULL, NULL, hInstance, NULL);
	if (!hwnd)
		return 0;
	hmenu = GetMenu (hwnd);
	
	ShowWindow (hwnd, iCmdShow);
	UpdateWindow (hwnd);
	
// main loop
	while (1)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			// Process the message
			if (msg.message == WM_QUIT)
				break;
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		else
		{
			cur_tick = GetTickCount();
			if (cur_tick > next_repaint_tick)
				if (game_state == GAME_RUNNING)
				{
					input();	
					physics();				
					repaint();
					next_repaint_tick = cur_tick + pause_between_frames;	
				}		
		}
	}
	return (int)msg.wParam;
}