#pragma once
#pragma comment(lib, "winmm.lib")

#include <vector>
#include <windows.h>
#include <tchar.h>
#include "resource.h"

using namespace std;

const char game_title[] = "Sixtis";

vector<int> tile;		// game field
vector<int> test_tile;	// copy of game field where we first copy the shape when trying to rotate, move or fall

double	shape_abs_x,
				shape_abs_y;
int shape_x,
		shape_y,		
		blocks_in_shape,
		shape_size;			// position and size of falling shape

int game_state,
		score;
		
double falling_speed (0.4),
			shifting_speed (5.0),
			fast_fall_speed (6.0);

enum {GAME_RUNNING, GAME_PAUSED};
enum {BORDER, EMPTY, BLOCK, SHAPE};

int	field_w (20),
		field_h (40), 
		field_size (800),
		tile_size (10),
		screen_width (200),
		screen_height (400), 
		pause_between_frames (20),
		max_blocks_in_shape (10), 
		min_blocks_in_shape (1),
		max_shape_size (7),
		score_for_rows[] = {0, 1, 3, 10, 30, 100, 300, 1000, 3000, 10000, 30000, 100000, 300000, 1000000};

int camera_x (0), camera_y (0); // position of the upper-left corner of the field on the screen 
																// (needed for auto-centering if field is too small)

HINSTANCE	hinstance;
HDC		hdc;
HWND	hwnd;
HMENU hmenu;
HBITMAP image_blocks, image_words;
vector<HBITMAP> loaded_images;

//--------------------------------------------------------------------------------------------------------------------
inline int random_int(const int& max)
{	
	if (max <= 1)	return 0;
	srand(GetTickCount() + rand());
	return (rand() % max);
}

//--------------------------------------------------------------------------------------------------------------------
inline void fetch_from_test_tile()
{
// clear the shape from tile vector and copy it from test_tile
	vector<int>::iterator	tile_it			= tile.begin(),
							test_tile_it	= test_tile.begin(),
							end_it			= tile.end();
	for(; tile_it != end_it; ++tile_it, ++test_tile_it)
	{
		if(*tile_it == SHAPE)			*tile_it = EMPTY;			
		if(*test_tile_it == SHAPE)		*tile_it = SHAPE;		
	}
}
//-----------------------------------------------------------------------------------------------------------------------
inline void clear_test_tiles()
{	
	vector<int> n (field_size, EMPTY);
	test_tile = n;
}

//--------------------------------------------------------------------------------------------------------------------
inline bool in_field(const int& a, const int& b){
	return(a >= 0 && a < field_w && b >= 0 && b < field_h);
}

//--------------------------------------------------------------------------------------------------------------------
inline HBITMAP loadImage (char* resource_name)
{	
	BITMAP bitmap;
	HBITMAP hBitmap;
	hBitmap = LoadBitmap (hinstance, TEXT (resource_name));
	GetObject (hBitmap, sizeof (BITMAP), &bitmap);
	loaded_images.push_back (hBitmap);	// remember that this image was loaded (will be freed when game exits)
	return hBitmap;
}

//--------------------------------------------------------------------------------------------------------
inline void repaint (bool erase_bkg = false)
{
	RECT r;
	r.right = 0;
	r.top = 0;
	r.left = screen_width;
	r.bottom = screen_height;
	InvalidateRect (hwnd, &r, erase_bkg);
}

//-----------------------------------------------------------------------------------------------------------
inline void drawImagePart (HBITMAP hBitmap, int x, int y, int sx, int sy, int w, int h)
{
	if (hBitmap != NULL)
	{
		HDC hdcMem = CreateCompatibleDC (hdc) ;
		SelectObject (hdcMem, hBitmap) ;
		BitBlt (hdc, x, y, w, h, hdcMem, sx, sy, SRCCOPY) ;
		DeleteDC (hdcMem) ;
	}
}

//-----------------------------------------------------------------------------------------------------------
inline void drawImage (HBITMAP hBitmap, int x, int y, int w, int h)
{
	drawImagePart (hBitmap, x, y, 0, 0, w, h);
}

//----------------------------------------------------------------------------------------------

inline void drawWhiteFilledRect (int x, int y, int w, int h, int r, int g, int b)
{
	HPEN lpen = CreatePen (PS_SOLID, 1, RGB (r, g, b));	
	SelectObject (hdc, lpen) ;
	Rectangle (hdc, x, y, x + w, y + h);
	DeleteObject (lpen);	
}

//-----------------------------------------------------------------------------------------------------------
int message (TCHAR * szFormat, ...)	
{
	TCHAR szBuffer [1024] ;
	va_list pArgList ;
	// The va_start macro (defined in STDARG.H) is usually equivalent to:
	// pArgList = (char *) &szFormat + sizeof (szFormat) ;
	va_start (pArgList, szFormat) ;
	// The last argument to wvsprintf points to the arguments
	_vsntprintf (szBuffer, sizeof (szBuffer) / sizeof (TCHAR),	szFormat, pArgList) ;
	// The va_end macro just zeroes out pArgList for no good reason
	va_end (pArgList) ;
	return MessageBox (NULL, szBuffer, "Sixtis", MB_OK);
}